@@ -4,6 +4,7 @@ from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                     List, Optional, overload, Sequence, Set, Tuple, Type,
                     TYPE_CHECKING, TypeVar, Union)
 import re
+import math
 import numpy as np 
 from pdb import set_trace
 import sympy
@@ -14,7 +15,7 @@ from cirq.circuits import InsertStrategy
 from cirq.protocols import qasm
 
 import quple
-from quple.components.gate_ops import RXX, RYY, RZZ
+from quple.components.gate_ops import RXX, RYY, RZZ, CompositeGate, PauliRotation
 from quple.circuits.qubit_register import QubitRegister
 from quple.utils.utils import parallel_run
 
@@ -31,7 +32,7 @@ kGateMapping = {
         "PauliZ": cirq.Z, # Pauli-Z gate
         "CX": cirq.CX, # Controlled-NOT gate
         "CNOT": cirq.CNOT, # Controlled-NOT gate
-        "CZ": cirq.CZ, # Controlled-Z gate
+        "CZ": cirq.CZ, # Controlled-Z ef
         "XX": cirq.XX, # tensor product of two X gates (X parity gate)
         "YY": cirq.YY, # tensor product of two Y gates (Y parity gate)
         "ZZ": cirq.ZZ, # tensor product of two Z gates (Z parity gate)
@@ -43,8 +44,8 @@ kGateMapping = {
         "ZZPowGate": cirq.ZZPowGate, # Z parity gate raised to some power
         "MS": cirq.ms, # Mølmer–Sørensen gate == RXX (A rotation around the XX axis in the two-qubit bloch sphere)
         "RXX": RXX, # XX Ising coupling gate (A rotation around the XX axis in the two-qubit bloch sphere)
-        "RXX": RYY, # YY Ising coupling gate (A rotation around the YY axis in the two-qubit bloch sphere)
-        "RXX": RZZ, # ZZ Ising coupling gate (A rotation around the ZZ axis in the two-qubit bloch sphere)                        
+        "RYY": RYY, # YY Ising coupling gate (A rotation around the YY axis in the two-qubit bloch sphere)
+        "RZZ": RZZ, # ZZ Ising coupling gate (A rotation around the ZZ axis in the two-qubit bloch sphere)                        
         "RX": cirq.rx, # rotation along X axis
         "RY": cirq.ry, # rotation along Y axis
         "RZ": cirq.rz, # rotation along Z axis
@@ -60,6 +61,7 @@ kGateMapping = {
         "CXPowGate": cirq.CXPowGate, # Controlled Power of an X gate
         "CZPowGate": cirq.CZPowGate, # Controlled Power of an Z gate
         "CNOTPowGate": cirq.CXPowGate, # Controlled Power of an X gate
+        "PauliRotation": PauliRotation # General Pauli Rotation 
     }
 
 
@@ -154,8 +156,8 @@ class QuantumCircuit(cirq.Circuit):
                 The insertion strategy of gate operations in the circuit.
                 If None, defaults to INLINE. 
             backend: default None
-                The backend for the quantum circuit. 
-                If None, defaults to quantum simulator.
+                The backend for circuit simulation. 
+                If None, defaults to the native TensorFlow simulator
         """
         super().__init__()
         self._name = name
@@ -178,7 +180,7 @@ class QuantumCircuit(cirq.Circuit):
     
 
     @n_qubit.setter
-    def n_qubit(self, value:int):
+    def n_qubit(self, value:Union[int, Sequence[GridQubit]]):
         self._qr = QubitRegister(value)
     
     @property
@@ -207,7 +209,7 @@ class QuantumCircuit(cirq.Circuit):
     
     @staticmethod
     def _parse_gate_operation(gate:Union[str, cirq.Gate]):
-        if isinstance(gate, cirq.Gate):
+        if isinstance(gate, (cirq.Gate, CompositeGate)):
             return gate
         elif isinstance(gate, str) and (gate in kGateMapping):
             return kGateMapping[gate]
@@ -221,8 +223,13 @@ class QuantumCircuit(cirq.Circuit):
     @property
     def raw_symbols(self):
         return quple.get_circuit_unflattened_symbols(self, to_str=True)  
+    
+    def set_insert_strategy(self, insert_strategy: InsertStrategy):
+        if not isinstance(insert_strategy, InsertStrategy):
+            raise ValueError('insert strategy must be derived from the cirq.circuits.InsertStrategy class')
+        self._insert_strategy = insert_strategy
 
-    def apply_gate_operation(self, operation:[str, cirq.Gate], qubit_expr, params=None):
+    def apply_gate_operation(self, operation:[str, cirq.Gate], qubit_expr, params:Optional[Dict]=None):
         """Apply a gate operation on designated qubits
         
         Args:
@@ -321,6 +328,33 @@ class QuantumCircuit(cirq.Circuit):
         """        
         self.apply_gate_operation(cirq.ops.ry(theta), qubit_expr)
 
+    def RXX(self, theta:Union[int, float], qubit_expr):
+        """Rotation about the X \otimes X axis
+        
+        RXX
+        
+        Performs two qubit rotation about the X \otimes X axis
+        """        
+        self.apply_gate_operation(RXX(theta), qubit_expr)  
+
+    def RXX(self, theta:Union[int, float], qubit_expr):
+        """Rotation about the Y \otimes Y axis
+        
+        RXX
+        
+        Performs two qubit rotation about the Y \otimes Y axis
+        """        
+        self.apply_gate_operation(RYY(theta), qubit_expr)     
+
+    def RZZ(self, theta:Union[int, float], qubit_expr):
+        """Rotation about the Z \otimes Z axis
+        
+        RXX
+        
+        Performs two qubit rotation about the Z \otimes Z axis
+        """        
+        self.apply_gate_operation(RZZ(theta), qubit_expr)                           
+
     def RZ(self, theta:Union[int, float], qubit_expr):
         """Rotation about the Z axis
         
@@ -328,7 +362,10 @@ class QuantumCircuit(cirq.Circuit):
         
         Performs single qubit rotation about the Z axis
         """        
-        self.apply_gate_operation(cirq.ops.rz(theta), qubit_expr)                
+        self.apply_gate_operation(cirq.ops.rz(theta), qubit_expr) 
+
+    def PauliRotation(self, pauli_string, theta:Union[int, float], qubit_expr, global_shift=False):
+        self.apply_gate_operation(PauliRotation(pauli_string, theta, global_shift=global_shift), qubit_expr)               
     
     def SWAP(self, qubit_expr):
         """The SWAP gate
@@ -544,7 +581,7 @@ class QuantumCircuit(cirq.Circuit):
                         the measurement is performed
             key: The string key of the measurement.
         """
-        qubit = self.qr.get(qubit_idx)
+        qubit = self.qr.get(qubit_expr)
         self.append(cirq.measure(qubit, key=key))
 
     def gate_operation(self, operation, qubit_expr, params=None):
@@ -583,7 +620,6 @@ class QuantumCircuit(cirq.Circuit):
         qubits = quple.get_circuit_qubits(circuit)
         symbols = quple.get_circuit_symbols(circuit)
         cq = cls(qubits)
-        cq._parameter_table.append(symbols)
         cq.append(circuit)
         return cq
 
@@ -631,15 +667,45 @@ class QuantumCircuit(cirq.Circuit):
         Returns:
             A circuit or list of circuits with parameter expressions resolved
         """
+        num_parameters = len(self.raw_symbols) 
+        if num_parameters == 0:
+            raise ValueError('circuit is not parameterised')
         param_resolver = self.get_parameter_resolver(vals)
         
         if isinstance(param_resolver, (list, cirq.ListSweep)):
             return parallel_run(self.get_resolved_circuit, param_resolver)
         else:
             return self.get_resolved_circuit(param_resolver)
+    
+    def resolve_parameters_with_tfq(self, vals:np.ndarray)-> Union['cirq.Circuit', List['cirq.Circuit']]:
+        import tensorflow_quantum as tfq
+        symbols, resolved_values = self._get_resolved_symbols_and_values(vals)
+        tensor_circuits =  tfq.convert_to_tensor([self for _ in range(len(resolved_values))])
+        resolved_circuits = tfq.from_tensor(tfq.resolve_parameters(tensor_circuits, symbols, resolved_values))
+        return resolved_circuits
                                 
     def get_resolved_circuit(self, param_resolver):
-        return cirq.protocols.resolve_parameters(self, param_resolver)        
+        return cirq.protocols.resolve_parameters(self, param_resolver)  
+    
+    def get_kernel_matrix_with_tfq(self, x_left:np.ndarray, x_right:np.ndarray=None, stride:int=50):
+        import tensorflow_quantum as tfq
+        import tensorflow as tf
+        resolved_circuits_left = tfq.convert_to_tensor(self.resolve_parameters(x_left))
+        if x_right == None:
+            resolved_circuits_right = resolved_circuits_left
+        else:
+            resolved_circuits_right = tfq.convert_to_tensor(self.resolve_parameters(x_right))
+        dimension = tf.gather(tf.shape(resolved_circuits_left), 0)
+        empty_symbols = tf.convert_to_tensor([], dtype=tf.dtypes.string)
+        empty_values = tf.tile(tf.convert_to_tensor([[]]), [stride, 1])
+        resolved_circuits_right = tf.reshape(tf.tile(resolved_circuits_right, [stride]), [stride, dimension])
+        
+        kernel_matrix = np.empty((dimension, dimension))
+        for i in range(0, dimension, stride):
+            interim = tfq.math.inner_product(resolved_circuits_left[i:i+stride], empty_symbols, empty_values, resolved_circuits_right)
+            kernel_matrix[i:i+stride] = (tf.math.abs(interim) ** 2).numpy()
+
+        return kernel_matrix
         
     def get_parameter_resolver(self, vals:np.ndarray) -> Union['cirq.ParamResolver',List['cirq.ParamResolver']]:
         """Returns a parameter resolver for the parameter expressions in the circuit
@@ -668,7 +734,7 @@ class QuantumCircuit(cirq.Circuit):
                 raise ValueError('Parameter values of type list must have elements '
                                  'of dict type representing the map from sympy.Symbol '
                                  'objects to symbol values')
-        elif isinstance(vals, np.ndarray):
+        elif isinstance(vals, np.ndarray) or hasattr(vals, 'ndim'):
             ndim = vals.ndim
             if ndim == 1:
                 vals = vals.reshape(-1,vals.size)
@@ -695,7 +761,45 @@ class QuantumCircuit(cirq.Circuit):
         symbols = list(resolvers[0].param_dict.keys())
         return symbols, values
     
+    def _get_tfq_layer_result(self, layer, vals:Optional[np.ndarray]=None, **kwargs):
+        num_parameters = len(self.raw_symbols) 
+        if num_parameters > 0:
+            if vals is None:
+                raise ValueError('Parameter values are required for a parameterized circuit')    
+            ndim = vals.ndim
+            if ndim != 2:
+                raise ValueError('Parameter values must be a 2D array')
+            if num_parameters != vals.shape[1]:
+                raise ValueError('Parameter dimension does not match the number of symbols in parameterized circuit')
+            symbols, resolved_values = self._get_resolved_symbols_and_values(vals)
+            result = layer(self, symbol_names=symbols, symbol_values=resolved_values, **kwargs)
+        else:
+            if vals is not None:
+                raise ValueError('Parameter values not allowed for non-parameterized circuit')
+            result = layer(self, **kwargs)
+        return result
+    
     def get_state_vectors(self, vals:Optional[np.ndarray]=None):
+        import tensorflow_quantum as tfq
+        state_layer = tfq.layers.State(backend=self.backend)
+        states = self._get_tfq_layer_result(state_layer, vals)
+        state_vectors = np.array([state.numpy() for state in states])
+        return state_vectors
+    
+    def get_expectation(self, vals:Optional[np.ndarray]=None, operators=None, repetitions=5000):
+        '''Sample expectation values given measurement operators and repetitions
+        '''
+        if operators is None:
+            operators = [cirq.Z(self.qubits[0])]
+        import tensorflow_quantum as tfq
+        expactation_layer = tfq.layers.SampledExpectation(backend=self.backend)
+        results = self._get_tfq_layer_result(expactation_layer, vals,
+                                             operators=operators,
+                                             repetitions=repetitions)
+        expectation = np.array([result.numpy() for result in results])
+        return expectation
+    
+    def get_state_vectors_deprecated(self, vals:Optional[np.ndarray]=None):
         import tensorflow_quantum as tfq
         state_layer = tfq.layers.State(backend=self.backend)
         num_parameters = len(self.raw_symbols) 
@@ -714,6 +818,34 @@ class QuantumCircuit(cirq.Circuit):
                 raise ValueError('Parameter values not allowed for non-parameterized circuit')
             states = state_layer(self)
         state_vectors = np.array([state.numpy() for state in states])
+        return state_vectors    
+    
+    def get_state_vectors_test(self, x:Optional[np.ndarray]=None, max_dim = 2000):
+        import tensorflow_quantum as tfq
+        state_layer = tfq.layers.State(backend=self.backend)
+        num_parameters = len(self.raw_symbols) 
+        if num_parameters > 0:
+            if x is None:
+                raise ValueError('Parameter values are required for a parameterized circuit')    
+            ndim = x.ndim
+            if ndim != 2:
+                raise ValueError('Parameter values must be a 2D array')
+            if num_parameters != x.shape[1]:
+                raise ValueError('Parameter dimension does not match the number of symbols in parameterized circuit')
+            symbols, resolved_values = self._get_resolved_symbols_and_values(x)
+            dim = len(resolved_values)
+            state_vectors = np.zeros((dim, 2**self.n_qubit), dtype=np.complex64)
+            for i, symbol_values in enumerate(np.array_split(resolved_values, math.ceil(dim/max_dim))):
+                split_dim = len(symbol_values)
+                states = state_layer(self, symbol_names=symbols, symbol_values=symbol_values)
+                for j, state in enumerate(states):
+                    state_vectors[i*max_dim+j] = state.numpy().copy()
+                    state = None
+        else:
+            if x is not None:
+                raise ValueError('Parameter values not allowed for non-parameterized circuit')
+            states = state_layer(self)
+            state_vectors = np.array([state.numpy() for state in states])
         return state_vectors
     
     def to_qasm_str(self):
