@@ -94,7 +94,6 @@ def get_circuit_unflattened_symbols(circuit:'quple.QuantumCircuit',
                                             sort_key=sort_key)
     return symbols
 
-
 #reference: https://github.com/tensorflow/quantum/blob/v0.3.0/tensorflow_quantum/python/util.py
 def get_circuit_symbols(circuit, to_str=True, sort_key=natural_key):
     """Returns a list of parameter symbols in a circuit
@@ -176,17 +175,12 @@ def sample_final_states(circuit, samples=1, data=None, backend=None):
     Returns:
         A list of sampled final states of the circuit
     """      
-    import tensorflow_quantum as tfq
-    layer = tfq.layers.State(backend=backend)
-    symbols = get_circuit_symbols(circuit)
-    if symbols:
-        if data is not None:
-            data = np.random.rand(samples, len(symbols))*2*np.pi
-        tensor_final_states = layer(circuit, symbol_names=symbols, symbol_values=data)
-    else:
-        tensor_final_states = layer([circuit]*samples)
-    final_states = [tensor_state.numpy() for tensor_state in tensor_final_states]
-    return final_states
+    if isinstance(circuit, cirq.Circuit) and not isinstance(circuit, quple.QuantumCircuit):
+        circuit = QuantumCircuit.from_cirq(circuit)
+    n_symbols = len(circuit.raw_symbols)
+    if (n_symbols > 0) and (data is None):
+        data = np.random.rand(samples, n_symbols)*2*np.pi
+    return circuit.get_state_vectors(data)
 
 def sample_density_matrices(circuit, samples=1, data=None, backend=None):
     """Samples the density matrices of a circuit
@@ -532,3 +526,26 @@ def gradient_variance_test(circuits, op, symbol=None):
     grad_var = tf.math.reduce_std(grads, axis=0)
     return grad_var.numpy()[0]
 
+
+def has_composite_symbols(circuit):
+    if not isinstance(circuit, cirq.Circuit):
+        raise ValueError("circuit must be a cirq.Circuit object.")
+    if (isinstance(circuit, quple.QuantumCircuit)) and (circuit.expr_map is not None):
+        circuit = circuit.get_unflattened_circuit()     
+    for moment in circuit:
+        for op in moment:
+            if cirq.is_parameterized(op):
+                if len(symbols_in_op(op.gate)) > 1:
+                    return True
+    return False
+'''
+    if not isinstance(circuit, quple.QuantumCircuit):
+        circuit = quple.QuantumCircuit.from_cirq(circuit)
+    # circuit is already flattened
+    if circuit.expr_map is not None:
+        return set(circuit.symbols) != set(circuit.raw_symbols)
+    circuit_symbols = circuit.symbols
+    flattened_circuit, _ = cirq.flatten(circuit)
+    flattened_circuit_symbols = get_circuit_symbols(flattened_circuit)
+    return set(circuit_symbols) != set(flattened_circuit_symbols)
+'''
\ No newline at end of file
